/*
  Copyright (c) 2022  Kazuhiro Ito (kzhr@d1.dion.ne.jp)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <ebu/eb.h>
#include <ebu/text.h>
#include <ebu/font.h>
#include <ebu/appendix.h>
#include <ebu/error.h>
#include <ebu/binary.h>
#include <cabin.h>
#include <estraier.h>

#include "common.h"

#define EBEST_INDEX_NAME "_ebest"

const char *program_name = "ebestmk";
const char *program_version = VERSION;

KVS *variables;
FILE *debug = NULL;
FILE *verbose = NULL;
bool signaled = false;

static const char *short_options = "g:s:hVvo";
static struct option long_options[] = {
  {"gaiji-map",         required_argument, NULL, 'g'},
  {"set",               required_argument, NULL, 's'},
  {"debug",             no_argument,       NULL, 'd'},
  {"help",              no_argument,       NULL, 'h'},
  {"variables",         no_argument,       NULL, 'X'},
  {"verbose",           no_argument,       NULL, 'V'},
  {"version",           no_argument,       NULL, 'v'},
  {"no-optimize",       no_argument,       NULL, 'o'},
  {"onp",               no_argument,       NULL, 'P'},
  {"ond",               no_argument,       NULL, 'D'},
  {"tr",                no_argument,       NULL, '1'},
  {EBEST_DB_TRUNC,      no_argument,       NULL, '1'},
  {"apn",               no_argument,       NULL, '2'},
  {EBEST_DB_PERFNG,     no_argument,       NULL, '2'},
  {"acc",               no_argument,       NULL, '3'},
  {EBEST_DB_CHRCAT,     no_argument,       NULL, '3'},
  {"xs",                no_argument,       NULL, '4'},
  {EBEST_DB_SMALL,      no_argument,       NULL, '4'},
  {"xl",                no_argument,       NULL, '5'},
  {EBEST_DB_LARGE,      no_argument,       NULL, '5'},
  {"xh",                no_argument,       NULL, '6'},
  {EBEST_DB_HUGE,       no_argument,       NULL, '6'},
  {"xh2",               no_argument,       NULL, '7'},
  {EBEST_DB_HUGE2,      no_argument,       NULL, '7'},
  {"xh3",               no_argument,       NULL, '8'},
  {EBEST_DB_HUGE3,      no_argument,       NULL, '8'},
  {"cs",                required_argument, NULL, '9'},
  {EBEST_DB_CACHE,      required_argument, NULL, '9'},
  {NULL,                no_argument,       NULL, 0}
};

const char *invoked_name;

static void show_version ()
{
  char buffer[256];

  printf ("%s %s (with EBU %d.%d, libestraier %s)\n",
	  program_name, program_version,
	  EB_VERSION_MAJOR, EB_VERSION_MINOR, _EST_VERSION);

  puts ("Copyright (C) 2022 Kazuhiro Ito");
  puts ("ebestmk may be distributed under the terms of the GNU General Public Licence;");
  puts ("certain other uses are permitted as well.  For details, see the file");
  puts ("`COPYING'.");
  puts ("There is no warranty, to the extent permitted by law.");
}

static void show_help ()
{
  fprintf (stderr, "Usage: %s [options...] [book-directory [appendix-directory]]\n", program_name);
  fprintf (stderr, "Options:\n");
  fprintf (stderr, "  -s, --set NAME[=VAL]       [un]set internal variable NAME [to VAL]\n");
  fprintf (stderr, "    variables:\n");
  fprintf (stderr, "        estdb_trunc          on|off  remove existing DB\n");
  fprintf (stderr, "        estdb_perfng         on|off  create DB with ESTDBPERFNG\n");
  fprintf (stderr, "        estdb_chrcat         on|off  create DB with ESTDBCHRCAT\n");
  fprintf (stderr, "        estdb_small          on|off  create DB with ESTDBSMALL\n");
  fprintf (stderr, "        estdb_large          on|off  create DB with ESTDBLARGE\n");
  fprintf (stderr, "        estdb_huge           on|off  create DB with ESTDBHUGE\n");
  fprintf (stderr, "        estdb_huge2          on|off  create DB with ESTDBHUGE2\n");
  fprintf (stderr, "        estdb_huge3          on|off  create DB with ESTDBHUGE3\n");
  fprintf (stderr, "        estdb_optimize       on|off  optimize DB before finish\n");
  fprintf (stderr, "        estdb_optimize_purge on|off  purge deleted content in optimizing\n");
  fprintf (stderr, "        estdb_optimize_db    on|off  optimize DB files in optimizing\n");
  fprintf (stderr, "        estdb_cache          decimal DB cache size in MiB\n");
  fprintf (stderr, "        ebu_gaiji_alternate  str     string for unallocated gaiji\n");
  fprintf (stderr, "        ebu_use_gaiji_map    on|off  use (or not use) gaiji replacement map\n");
  fprintf (stderr, "        ebu_coding_epwing    str     iconv name for EPWING\n");
  fprintf (stderr, "        ebu_coding_ebxac     str     iconv name for Chinese part of EBXA-C\n");
  fprintf (stderr, "        ebu_coding_ebg       str     iconv name for EBG\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "  -g, --gaiji-map on|off  abbr. for --set ebu_use_gaiji_map=on|off\n");
  fprintf (stderr, "  -o, --no-optimize       abbr. for --set estdb_optimize=off\n");
  fprintf (stderr, "      --ond               abbr. for --set estdb_optimize_db=off\n");
  fprintf (stderr, "      --onp               abbr. for --set estdb_optimize_purge=off\n");
  fprintf (stderr, "      --tr                abbr. for --set estdb_trunc=on\n");
  fprintf (stderr, "      --apn               abbr. for --set estdb_perfng=on\n");
  fprintf (stderr, "      --acc               abbr. for --set estdb_chrcat=on\n");
  fprintf (stderr, "      --xs                abbr. for --set estdb_small=on\n");
  fprintf (stderr, "      --xl                abbr. for --set estdb_large=on\n");
  fprintf (stderr, "      --xh                abbr. for --set estdb_huge=on\n");
  fprintf (stderr, "      --xh2               abbr. for --set estdb_huge2=on\n");
  fprintf (stderr, "      --xh3               abbr. for --set estdb_huge3=on\n");
  fprintf (stderr, "      --cs  decimal       abbr. for --set estdb_cache=decimal\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "      --variables          show internal variables\n");
  fprintf (stderr, "  -V, --verbose            verbose mode\n");
  fprintf (stderr, "  -d, --debug              show debug messages\n");
  fprintf (stderr, "  -v, --version            show version number\n");
  fprintf (stderr, "  -h, --help               show this message\n");
  fflush (stderr);
}

static void show_try_help ()
{
  fprintf (stderr, "Try `%s --help' for more information.\n", invoked_name);
  fflush (stderr);
}

static void show_variables() {
  printf("Internal variables\n");
  for (int i = 0; i < variables->size; i++) {
    if (variables->list[i].value) {
      if (variables->list[i].key[0] != '_') {
	printf("%s=%s\n", variables->list[i].key, variables->list[i].value);
      } else {
	printf("%s=%s\n", variables->list[i].key + 1,
	       variables->list[i].value);
      }
    }
  }
}

static int init_variables() {
  variables = kvs_create();

  variables_set(variables, "name", program_name, true);
  variables_set(variables, "version", VERSION, true);
  variables_set
    (variables, EBU_CODING_EPWING, EBU_CODING_EPWING_DEFAULT, false);
  variables_set
    (variables, EBU_CODING_EBXAC , EBU_CODING_EBXAC_DEFAULT , false);
  variables_set
    (variables, EBU_CODING_EBG   , EBU_CODING_EBG_DEFAULT   , false);
  variables_set
    (variables, EBU_USE_GAIJI_MAP, "on"                     , false);
  variables_set
    (variables, EBU_GAIJI_ALTERNATE, EBU_GAIJI_ALTERNATE_DEFAULT, false);
  variables_set(variables, EBEST_DB_CACHE, EBEST_DB_CACHE_DEFAULT, false);
  variables_set(variables, EBEST_DB_TRUNC,     "off", false);
  variables_set(variables, EBEST_DB_PERFNG,    "off", false);
  variables_set(variables, EBEST_DB_CHRCAT,    "off", false);
  variables_set(variables, EBEST_DB_SMALL,     "off", false);
  variables_set(variables, EBEST_DB_LARGE,     "off", false);
  variables_set(variables, EBEST_DB_HUGE,      "off", false);
  variables_set(variables, EBEST_DB_HUGE2,     "off", false);
  variables_set(variables, EBEST_DB_HUGE3,     "off", false);
  variables_set(variables, EBEST_DB_OPTIMIZE,  "on",  false);
  variables_set(variables, EBEST_DB_OPT_PURGE, "on", false);
  variables_set(variables, EBEST_DB_OPT_DB,    "on",  false);
  
  return 0;
}

static void estdb_set_cache_size (ESTDB *estdb) {
  size_t csize;
  const char *val =
    variables_get(variables, EBEST_DB_CACHE, EBEST_DB_CACHE_DEFAULT);
  char *endptr;
  csize = strtol(val, &endptr, 10);
  if (endptr == val || *endptr != 0) {
    csize = strtol(EBEST_DB_CACHE_DEFAULT, &endptr, 10);
  }

  if (csize > ((size_t) -1) / (1024 * 1024)) {
    csize = (size_t) -1;
  } else {
    csize *= (1024 * 1024);
  }
  VERBOSE("estdb cache size is %dMiB", ((csize + 512) / 1024 + 512) / 1024);

  int r = csize / (64 * 1024 * 1024);
  if  (r < 1) r = 1;
  else if (r > 1000) r = 1000;
  
  est_db_set_cache_size(estdb, csize, 8192 * r, 1024 * r, 256 * r);
}


void ebestmk_signal_handler (int signum) {
  const char *name = NULL;

  switch(signum) {
#ifdef SIGABRT
  case SIGABRT:
    name = "SIGABRT";
    break;
#endif
#ifdef SIGHUP
  case SIGHUP:
    name = "SIGHUP";
    break;
#endif
#ifdef SIGINT
  case SIGINT:
    name = "SIGINT";
    break;
#endif
#ifdef SIGQUIT
  case SIGQUIT:
    name = "SIGQUIT";
    break;
#endif
#ifdef SIGTERM
  case SIGTERM:
    name = "SIGTERM";
    break;
#endif
  default:
    break;
  }

  if (name) {
    printf("Caught %s!\n", name);
  } else {
    /* should not reach */
    printf("Caught signal %d!\n", signum);
  }
  signaled = true;

#ifndef HAVE_SIGACTION
  signal(signum, SIG_DFL);
#endif
  return;
}

#ifdef HAVE_SIGACTION
void set_signal_handler() {
  struct sigaction action;

  action.sa_handler = ebestmk_signal_handler;
  action.sa_flags = SA_RESETHAND;
  sigfillset(&(action.sa_mask));

#ifdef SIGABRT
  sigaction(SIGABRT, &action, NULL);
#endif
#ifdef SIGHUP
  sigaction(SIGHUP, &action, NULL);
#endif
#ifdef SIGINT
  sigaction(SIGINT, &action, NULL);
#endif
#ifdef SIGQUIT
  sigaction(SIGQUIT, &action, NULL);
#endif
#ifdef SIGTERM
  sigaction(SIGTERM, &action, NULL);
#endif
}
#else
void set_signal_handler() {
#ifdef SIGABRT
  signal(SIGABRT, ebestmk_signal_handler);
#endif
#ifdef SIGHUP
  signal(SIGHUP, ebestmk_signal_handler);
#endif
#ifdef SIGINT
  signal(SIGINT, ebestmk_signal_handler);
#endif
#ifdef SIGQUIT
  signal(SIGQUIT, ebestmk_signal_handler);
#endif
#ifdef SIGTERM
  signal(SIGTERM, ebestmk_signal_handler);
#endif
}
#endif

int main (int argc, char** argv) {
  bool flag_variables = false;
  invoked_name = argv[0];

  if (init_variables()) {
    fprintf(stderr, "Failed to initialize internal variables\n");
    return -1;
  }

  /* parse command line options */
  int optch;
  while ((optch = getopt_long(argc, argv, short_options, long_options, NULL))
	 != EOF) {
    switch (optch) {
    case 't':
      break;
    case 'd':
      variables_set (variables, EBEST_DEBUG, "on", false);
      break;
    case 'g':
      variables_set (variables, EBU_USE_GAIJI_MAP, optarg, false);
      break;
    case 's':
      variagles_set_from_line(variables, optarg);
      break;
    case 'h':
      show_help ();
      exit (0);
    case 'v':
      show_version ();
      exit (0);
    case 'V':
      variables_set(variables, EBEST_VERBOSE, "on", false);
      break;
    case 'X':
      flag_variables = true;
      break;
    case 'o':
      variables_set(variables, EBEST_DB_OPTIMIZE,  "off", false);
      break;
    case 'D':
      variables_set(variables, EBEST_DB_OPT_DB,    "off", false);
      break;
    case 'P':
      variables_set(variables, EBEST_DB_OPT_PURGE, "off", false);
      break;
    case '1':
      variables_set(variables, EBEST_DB_TRUNC,     "on", false);
      break;
    case '2':
      variables_set(variables, EBEST_DB_PERFNG,    "on", false);
      break;
    case '3':
      variables_set(variables, EBEST_DB_CHRCAT,    "on", false);
      break;
    case '4':
      variables_set(variables, EBEST_DB_SMALL,     "on", false);
      break;
    case '5':
      variables_set(variables, EBEST_DB_LARGE,     "on", false);
      break;
    case '6':
      variables_set(variables, EBEST_DB_HUGE,      "on", false);
      break;
    case '7':
      variables_set(variables, EBEST_DB_HUGE2,     "on", false);
      break;
    case '8':
      variables_set(variables, EBEST_DB_HUGE3,     "off", false);
      break;
    case '9':
      variables_set(variables, EBEST_DB_CACHE,     optarg, false);
      break;
    default:
      show_try_help ();
      exit (1);
    }
  }

  if (variables_get_bool(variables, EBEST_DEBUG, false)) {
    debug = stderr;
    verbose = stderr;
  } else if (variables_get_bool(variables, EBEST_VERBOSE, false)) {
    verbose = stdout;
  }

  if (flag_variables) {
    show_variables();
    exit (0);
  }

  /* check the rest arguments */
  const char *path = NULL;
  switch (argc - optind) {
  case 0:
    fprintf (stderr, "%s: too few arguments\n", invoked_name);
    show_try_help ();
    exit (1);
  }

  path = argv[optind];

  EBU_Params ebu;
  if (ebu_open (argc, argv, optind, &ebu)) {
    return -1;
  }

  char* estdb_path;
  if (path[strlen(path) - 1] == '/') {
    estdb_path = snprintf_buffer("%s" EBEST_INDEX_NAME, path);
  } else {
    estdb_path = snprintf_buffer("%s/" EBEST_INDEX_NAME, path);
  }
  int est_error;

  ESTDB *estdb =
    est_db_open(estdb_path,
		ESTDBWRITER | ESTDBCREAT |
		(variables_get_bool(variables, EBEST_DB_TRUNC, false)
		 ? ESTDBTRUNC : 0) |
		(variables_get_bool(variables, EBEST_DB_SMALL, false)
		 ? ESTDBSMALL : 0) |
		(variables_get_bool(variables, EBEST_DB_LARGE, false)
		 ? ESTDBLARGE : 0) |
		(variables_get_bool(variables, EBEST_DB_HUGE,  false)
		 ? ESTDBHUGE  : 0) |
		(variables_get_bool(variables, EBEST_DB_HUGE2, false)
		 ? ESTDBHUGE2 : 0) |
		(variables_get_bool(variables, EBEST_DB_HUGE3, false)
		 ? ESTDBHUGE3 : 0) |
		(variables_get_bool(variables, EBEST_DB_PERFNG, false)
		 ? ESTDBPERFNG : 0) |
		(variables_get_bool(variables, EBEST_DB_CHRCAT, false)
		 ? ESTDBCHRCAT : 0),
		&est_error);
  if (!estdb) {
    fprintf(stderr, "Failed to open database %s\n", estdb_path);
    return -1;
  }
  free(estdb_path);
  ebu.estdb = estdb;

  estdb_set_cache_size(estdb);

  est_db_add_attr_index(estdb, "@uri", ESTIDXATTRSTR);
  est_db_add_attr_index(estdb, "@title", ESTIDXATTRSTR);
  if (verbose) {
    est_db_set_informer(estdb, estdb_informer, NULL);
  }

  ebu_make_index(&ebu);
  
  if (variables_get_bool(variables, EBEST_DB_OPTIMIZE, true) && !signaled) {
    puts("optimizing...");
    int opt = 0;
    if (!variables_get_bool(variables, EBEST_DB_OPT_PURGE, false)
	|| estdb->idxdb->dnum == 1)
      opt |= ESTOPTNOPURGE;
    if (!variables_get_bool(variables, EBEST_DB_OPT_DB, true))
      opt |= ESTOPTNODBOPT;
    est_db_optimize(estdb, opt);
  }
  est_db_close(estdb, &est_error);

  return 0;
}
