/*
  Copyright (c) 2022  Kazuhiro Ito (kzhr@d1.dion.ne.jp)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <iconv.h>
#include <ebu/eb.h>
#include <ebu/text.h>
#include <ebu/font.h>
#include <ebu/appendix.h>
#include <ebu/error.h>
#include <ebu/binary.h>

#include "common.h"


static bool initialized = false;

#define BUFFER_START_SIZE (4096)
typedef struct {
  EB_Subbook_Code code;
  SBuffer *uribase;
  char *name;
  KVS *gaiji;
} EBUSUB;

typedef struct {
  EBU_Params *ebu;
  EBUSUB *ebusub;
  ESTDB *estdb;
  EB_Position pos;
  iconv_t iconv;
  iconv_t iconv_ebxac;
  SBuffer *ibuffer;
  SBuffer *hbuffer;
  SBuffer *obuffer;
  SBuffer *uri;
} EBU_Content;


/*
 * Text hooks
 */
EB_Error_Code hook_iso8859_1
(EB_Book *book, EB_Appendix *appendix, void *container, EB_Hook_Code code,
 int argc, const unsigned int *argv) {
  switch (argv[0]) {
  case '&':
    return eb_write_text_string(book, "&&;");
  }
  return eb_write_text_byte1(book, argv[0]);
}
  

EB_Error_Code hook_euc_to_ascii
(EB_Book *book, EB_Appendix *appendix, void *container, EB_Hook_Code code,
 int argc, const unsigned int *argv) {
  switch (argv[0]){
  case 0xa1f5: /* wide "&" */
    return eb_write_text_string(book, "&&;");
  }
  return eb_hook_euc_to_ascii (book, appendix, container, code, argc, argv);
}

EB_Error_Code hook_font
(EB_Book *book, EB_Appendix *appendix, void *container, EB_Hook_Code code,
 int argc, const unsigned int *argv) {
  char buff[EB_MAX_ALTERNATION_TEXT_LENGTH + 1];
  char gaiji_key[5 + 1];
  EB_Error_Code result = EB_SUCCESS;
  KVS *gaiji = container;

  switch (code) {
  case EB_HOOK_WIDE_JISX0208:
    if (argv[0] < 0xada0 || (argv[0] >= 0xb0a0 && argv[0] <= 0xf4af)) {
      return eb_write_text_byte2(book, argv[0] >> 8, argv[0] & 0xff);
    }
    sprintf(buff, "&e%04x;", argv[0]);
    break;
  case EB_HOOK_NARROW_FONT:
    sprintf(gaiji_key, "h%04x", argv[0]);
    if (gaiji && kvs_get(gaiji, gaiji_key)) {
      sprintf (buff, "&%s;", gaiji_key);
    } else if (!appendix ||
	       eb_narrow_alt_character_text (appendix, argv[0], buff)
	       != EB_SUCCESS) {
      sprintf (buff, "&%s;", gaiji_key);
    }
    break;
  case EB_HOOK_WIDE_FONT:
    return EB_SUCCESS;
    sprintf(gaiji_key, "z%04x", argv[0]);
    if (gaiji && kvs_get(gaiji, gaiji_key)) {
      sprintf (buff, "&%s;", gaiji_key);
    } else if (!appendix ||
	       eb_wide_alt_character_text (appendix, argv[0], buff)
	       != EB_SUCCESS) {
      sprintf (buff, "&%s;", gaiji_key);
    }
    break;
  case EB_HOOK_GB2312:
    sprintf (buff, "&g%04x;", argv[0]);
    break;
  case EB_HOOK_EBXAC_GAIJI:
    sprintf (buff, "&c%04x;", argv[0] & 0x7f7f);
    break;
  default:
    return EB_SUCCESS;
  }
  DEBUG("Using gaiji %s", buff);
  return eb_write_text_string(book, buff);
}

EB_Error_Code hook_unicode_tags
(EB_Book *book, EB_Appendix *appendix, void *container, EB_Hook_Code code,
 int argc, const unsigned int *argv) {
  switch (code) {
  case EB_HOOK_BEGIN_UNICODE:
    return eb_write_text_string(book, "&u");
    break;
  case EB_HOOK_END_UNICODE:
    return eb_write_text_string(book, ";");
    break;
  }
  return EB_SUCCESS;
}

EB_Hook text_hooks[] = {
  {EB_HOOK_ISO8859_1,       hook_iso8859_1},
  {EB_HOOK_NARROW_JISX0208, hook_euc_to_ascii},
  {EB_HOOK_NARROW_FONT,     hook_font},
  {EB_HOOK_WIDE_FONT,	    hook_font},
  {EB_HOOK_GB2312,	    hook_font},
  {EB_HOOK_EBXAC_GAIJI,	    hook_font},
  {EB_HOOK_WIDE_JISX0208,   hook_font},
  {EB_HOOK_BEGIN_UNICODE,   hook_unicode_tags},
  {EB_HOOK_END_UNICODE,     hook_unicode_tags},
  {EB_HOOK_NULL, NULL},
};

static void ebu_initialize() {
  if (!initialized) {
    EB_Error_Code error_code = eb_initialize_library ();
    if (error_code != EB_SUCCESS) {
      fprintf (stderr, "Failed to initialize EB library.\n");
      exit(1);
    }
    initialized = true;
  }
}

static void ebu_finalize() {
  if (initialized) {
    eb_finalize_library ();
    initialized = false;
  }
}

int ebu_open (int argc, char** argv, int start, EBU_Params *ebu) {
  /* check the rest arguments */
  const char *book = NULL, *appendix = NULL;
  switch (argc - start) {
  case 2:
    appendix = argv[start + 1];
  case 1:
    book = argv[start];
    break;
  case 0:
    fputs("Too few arguments.\n", stderr);
    goto error;
  default:
    fputs("Too many arguments.\n",stderr);
    goto error;
  }

  ebu_initialize();
  EB_Error_Code error_code;

  ebu->appendix = NULL;
  eb_initialize_hookset (&ebu->hookset);
  ebu->book = malloc(sizeof(EB_Book));
  if (!ebu->book) return 1;
  eb_initialize_book (ebu->book);
  ebu->gaiji_default = variables_get(variables, EBU_GAIJI_ALTERNATE, NULL);

  VERBOSE("ebu book is %s", book);
  error_code = eb_bind (ebu->book, book);
  if (EB_SUCCESS != error_code) {
    fprintf (stderr, "eb_bind(): %s\n", eb_error_message(error_code));
    goto error;
  }
  
  EB_Character_Code book_charcode;
  error_code = eb_character_code(ebu->book, &book_charcode);
  if (EB_SUCCESS != error_code) {
    fprintf (stderr, "eb_character_code(): %s\n",
	     eb_error_message(error_code));
    goto error;
  }
  const unsigned char *coding;
  switch(book_charcode) {
  case EB_CHARCODE_ISO8859_1:
    coding = variables_get(variables, EBU_CODING_EBG, EBU_CODING_EBG_DEFAULT);
  ebu->coding_ebxac = NULL;
  case EB_CHARCODE_JISX0208_GB2312:
    coding = variables_get(variables,
			   EBU_CODING_EBXAC, EBU_CODING_EBXAC_DEFAULT);
    ebu->coding_ebxac = coding;
  case EB_CHARCODE_JISX0208:
    coding = variables_get(variables,
			   EBU_CODING_EPWING, EBU_CODING_EPWING_DEFAULT);
    ebu->coding_ebxac = NULL;
    break;
  case EB_CHARCODE_UTF8:
    coding = NULL;
    ebu->coding_ebxac = NULL;
    break;
  }
  if (coding) VERBOSE("ebu book's coding is %s", coding);
  ebu->coding = coding;

  if (appendix) {
    VERBOSE("ebu appendix is %s", book);
    ebu->appendix = malloc(sizeof(EB_Appendix));
    if (!ebu->appendix) goto error;
    eb_initialize_appendix (ebu->appendix);
    error_code = eb_bind_appendix (ebu->appendix, appendix);
    if (error_code != EB_SUCCESS) {
      fprintf (stderr, "%s:eb_bind_appendix(): %s\n", appendix,
	       eb_error_message(error_code));
      goto error;
    }
  }

  error_code = eb_set_hooks (&ebu->hookset, text_hooks);
  if (error_code != EB_SUCCESS) {
    fprintf (stderr, "eb_set_hooks(): %s\n", eb_error_message(error_code));
    goto error;
  }
  
  return 0;

 error:
  if (ebu->book) ebu_close(ebu);
  return -1;
}

void ebu_close (EBU_Params *ebu) {
  if (ebu->book) {
    eb_finalize_book(ebu->book);
    free(ebu->book);
    ebu->book = NULL;
  }
  if (ebu->appendix) {
    eb_finalize_appendix(ebu->appendix);
    free(ebu->appendix);
    ebu->appendix = NULL;
  }
  eb_finalize_hookset (&ebu->hookset);
}

void ebu_content_destroy(EBU_Content *content) {
  if (content->uri) free(content->uri);
  if (content->iconv != (iconv_t) -1)
    iconv_close(content->iconv);
  if (content->iconv_ebxac != (iconv_t) -1)
    iconv_close(content->iconv_ebxac);
  if (!content->hbuffer) sbuffer_destroy(content->hbuffer);
  if (!content->ibuffer) sbuffer_destroy(content->ibuffer);
  if (!content->obuffer) sbuffer_destroy(content->obuffer);
  if (!content->uri)     sbuffer_destroy(content->uri);
  free(content);
}

void ebu_content_reset(EBU_Content *content) {
  content->ibuffer->used = 0;
  content->obuffer->used = 0;
  content->hbuffer->used = 0;
  content->uri->used = 0;
}

EBU_Content *ebu_content_create(EBU_Params *ebu, EBUSUB *ebusub) {
  EBU_Content *content = malloc(sizeof(EBU_Content));
  if (!content) goto fail;

  content->estdb = ebu->estdb;
  content->ebu = ebu;
  content->ebusub = ebusub;

  content->hbuffer = sbuffer_create(BUFFER_START_SIZE);
  content->ibuffer = sbuffer_create(BUFFER_START_SIZE);
  content->obuffer = sbuffer_create(BUFFER_START_SIZE);
  content->uri     = sbuffer_create(30);
  if (!(content->uri && content->hbuffer &&
	content->ibuffer && content->obuffer))
    goto fail;

  if (ebu->coding) {
    content->iconv = iconv_open("UTF-8", ebu->coding);
    if (content->iconv == (iconv_t) -1) {
      goto fail;
    }
  } else {
    content->iconv = (iconv_t) -1;
  }
  if (ebu->coding_ebxac) {
    content->iconv_ebxac = iconv_open("UTF-8", ebu->coding_ebxac);
    if (content->iconv_ebxac == (iconv_t) -1) {
      iconv_close(content->iconv);
      goto fail;
    }
  } else {
    content->iconv_ebxac = (iconv_t) -1;
  }

  return content;

 fail:
  if (content->hbuffer) sbuffer_destroy(content->hbuffer);
  if (content->ibuffer) sbuffer_destroy(content->ibuffer);
  if (content->obuffer) sbuffer_destroy(content->obuffer);
  if (content->uri)     sbuffer_destroy(content->uri);
  if (content) ebu_content_destroy(content);
  return NULL;
}

static EBUSUB *ebusub_create() {
  EBUSUB *result = malloc(sizeof(EBUSUB));
  if (!result) return NULL;

  result->uribase = sbuffer_create(30);;

  if (!(result->uribase)) {
    free(result);
    return NULL;
  }

  result->name = NULL;
  result->gaiji = NULL;

  return result;
}

static void ebusub_destroy(EBUSUB *ebusub) {
  if (ebusub->name) free(ebusub->name);
  if (ebusub->gaiji) kvs_destroy(ebusub->gaiji);
  if (ebusub->uribase) sbuffer_destroy(ebusub->uribase);

  free(ebusub);
}

static void ebu_set_content_title(EBU_Content *content) {
  SBuffer *obuffer = content->obuffer;
  SBuffer *hbuffer = content->hbuffer;

  size_t hlength = strchr(obuffer->data, '\n') - (char *)obuffer->data;
  sbuffer_request(hbuffer, hlength + 1);
  memcpy(hbuffer->data, obuffer->data, hlength);
  hbuffer->data[hlength] = 0;
  hbuffer->used = hlength + 1;
}

static int ebu_translate_content(EBU_Content *content) {
  KVS *gaiji = content->ebusub->gaiji;
  size_t pos = 0;
  unsigned char buffer[2 + 1];
  SBuffer *ibuffer = content->ibuffer;
  SBuffer *obuffer = content->obuffer;
  SBuffer *hbuffer = content->hbuffer;
  iconv_t iconvd = content->iconv;
  iconv_t iconv_ebxac = content->iconv_ebxac;
  int result = 0;

  while(pos < ibuffer->size) {
    char *ptr = strchr(ibuffer->data + pos, '&');
    if (!ptr) {
      DEBUG("content has no special cahacters");
      result = sbuffer_iconv(obuffer, iconvd,
			     ibuffer->data + pos,
			     ibuffer->used - pos);
      goto finish;
    } else {
      result = sbuffer_iconv(obuffer, iconvd, ibuffer->data + pos,
			     (ptr - (char *)ibuffer->data) - pos);
      if (result) return result;
      char *endp = strchr(ptr, ';');
      if (!endp) {
	fprintf(stderr, "ebu_translate_content(): invalid syntax\n");
	return -1;
      }

      int ch;
      char *p;
      const char *string;
      switch (*(ptr + 1)) {
      case 'z':
      case 'h':
      case 'c':
      case 'e':
	*endp = 0;
	string = gaiji ? kvs_get(gaiji, ptr + 1) : content->ebu->gaiji_default;
	DEBUG("GAIJI %s is replaced with %s",
	      ptr + 1, string ? string : "no string");

	if (string) sbuffer_write_bytes(obuffer, string, strlen(string));
	break;
      case 'g':
	*endp = 0;
	string = gaiji ? kvs_get(gaiji, ptr + 1) : NULL;
	if (string) {
	  DEBUG("GAIJI %s is replaced with %s",
		ptr + 1, string ? string : "no string");
	  sbuffer_write_bytes(obuffer, string, strlen(string));
	} else {
	  ch = (int) strtol (ptr + 2, &p, 16);
	  if (*p != 0) {
	    fprintf(stderr,
		    "ebu_translate_content(): not hex chars, %s\n", ptr + 1);
	    return -1;
	  }
	  buffer[0] = (ch >> 8) & 255;
	  buffer[1] =  ch       & 255;
	  result = sbuffer_iconv(obuffer, iconv_ebxac, buffer, 2);
	  if (result) return result;
	}
	break;
      case 'u':
	*endp = 0;
        ch = (int) strtol (ptr + 2, &p, 16);
	int j;
	if (*p != 0) {
	  ch = 0;
	  p = ptr + 2;
	  for (;*p != 0; p++) {
	    j = ((*p) << 8) + *(p + 1);
	    if (j >= 0xa3b0 && j <= 0xa3b9) {
	      ch = ch * 16 + j - 0xa3b0;
	    } else if (j >= 0xa3e1 && j <= 0xa3e6) {
	      ch = ch * 16 + j - 0xa3e1 + 10;
	    } else if (j >= 0xa3c1 && j <= 0xa3c6) {
	      ch = ch * 16 + j - 0xa3c1 + 10;
	    } else {
	      fprintf(stderr,
		      "ebu_translate_content(): not hex chars, %s\n", ptr + 1);
	      return -1;
	    }
	  }
	}

	sbuffer_write_utf8(obuffer, ch);
	break;
      case '&':
	sbuffer_write_bytes(obuffer, "&", 1);
	if (endp - ptr != 2) {
	  fprintf(stderr, "ebu_translate_content(): invalid syntax\n");
	  return -1;
	}
	break;
      }
      pos = endp - (char *)ibuffer->data  + 1;
    }
  }

 finish:
  return result;
}

static int ebu_read_content(EBU_Content *content) {
  EBU_Params *ebu = content->ebu;
  ssize_t length;
  SBuffer *ibuffer = content->ibuffer;

  do {
    sbuffer_request(ibuffer, BUFFER_START_SIZE);
    eb_read_text(ebu->book, ebu->appendix, &ebu->hookset,
		 content->ebusub->gaiji,
		 ibuffer->size - ibuffer->used - 1,
		 ibuffer->data + ibuffer->used, &length);
    ibuffer->used += length;
  } while (length != 0);
  DEBUG("content length is %d", ibuffer->used);

  if (ibuffer->used == 0) {
    fprintf(stderr, "uri %s has no content\n", content->uri->data);
    return 1;
  }
  sbuffer_write_char(ibuffer, 0);

  return 0;
}

static int ebu_register_content (EBU_Content *content) {
  const char *name = content->ebusub->name;
  KVS *gaiji = content->ebusub->gaiji;
  EBU_Params *ebu = content->ebu;
  ESTDB *estdb = ebu->estdb;
  EB_Book *book = ebu->book;
  int result = 0;
  EB_Position *pos = &(content->pos);

  ebu_content_reset(content);
  eb_tell_text(book, pos);

  if (sbuffer_write_snprintf (content->uri, content->ebusub->uribase->data,
			      pos->page, pos->offset) < 0) {
    result = -1;
    goto finish;
  }

  DEBUG("Making document %s", content->uri->data);

  if (est_db_uri_to_id(estdb, content->uri->data) != -1) {
    DEBUG("document %s is already registered", content->uri->data);
    result = 1;
    goto finish;
  }

  DEBUG("Reading document %s", content->uri->data);
  result = ebu_read_content(content);
  if (result) goto finish;

  DEBUG("Translating document %s", content->uri->data);
  if (ebu_translate_content(content)) {
    fprintf(stderr, "translation failed, skipped\n");
    result = -1;
    goto finish;
  }

  ebu_set_content_title(content);
  DEBUG("Translation done");

  ESTDOC* estdoc = est_doc_new();
  est_doc_add_attr(estdoc, "@title", content->hbuffer->data);
  est_doc_add_attr(estdoc, "@uri", content->uri->data);
  est_doc_add_text(estdoc, content->obuffer->data);
  VERBOSE("Registering %s (%s)", content->uri->data, content->hbuffer->data);
  est_db_put_doc(content->estdb, estdoc, 0);
  est_doc_delete(estdoc);

 finish:
  return result;
}

static void ebu_register_content0 (EBU_Params *ebu, EBUSUB *ebusub) {
  if (signaled) return;

  EBU_Content *content = ebu_content_create(ebu, ebusub);
  if (!content) return;

  int registered = 0, skipped = 0, failed = 0;

  do {
    switch (ebu_register_content(content)) {
    case 0:
      registered++;
      break;
    case 1:
      skipped++;
      break;
    case -1:
      failed++;
      break;
    default:
      break;
    }
  } while (eb_forward_text (ebu->book, NULL) == EB_SUCCESS && !signaled);

  printf("Registered %d, skipped %d, failed %d entries.\n",
	 registered, skipped, failed);

  ebu_content_destroy(content);
}

static int ebu_register_subbook_info (EBU_Params *ebu, EBUSUB *ebusub) {
  const char* name = ebusub->name;
  ESTDOC* estdoc = est_doc_new();
  char *str = snprintf_buffer("ebest:subinfo/%s", name);

  if (!str) goto error;
  est_doc_add_attr(estdoc, "@uri", str);
  free(str);
  if (!(str = snprintf_buffer("%d", ebusub->code + 1))) goto error;
  est_doc_add_attr(estdoc, "eblook_code", str);
  free(str);
  est_doc_add_attr(estdoc, "@title", "ebest subbook information");
  if (!(str = snprintf_buffer("%2d. %s\n", ebusub->code + 1, name)))
    goto error;
  est_doc_add_text(estdoc, str);
  free(str);
  DEBUG("Registering subbook information for %s", name);
  est_db_put_doc(ebu->estdb, estdoc, 0);
  DEBUG("Registered");
  est_doc_delete(estdoc);
  return 0;

 error:
  est_doc_delete(estdoc);
  return -1;
}

static int ebu_make_index_subbook (EBU_Params *ebu, EBUSUB *ebusub) {
  EB_Subbook_Code subbook_code = ebusub->code;
  const char* name = ebusub->name;
  printf ("Current subbook is %d.\n", subbook_code);

  if (ebu_register_subbook_info(ebu, ebusub)) {
    fprintf(stderr, "failed to register subbook information.\n");
  }

  EB_Error_Code error_code = eb_set_subbook(ebu->book, subbook_code);
  if (error_code != EB_SUCCESS) {
    fprintf (stderr, "eb_set_subbook: %s\n", eb_error_message(error_code));
    return 1;
  }

  if (ebu->appendix) {
    error_code = eb_set_appendix_subbook(ebu->appendix, subbook_code);
    if (error_code != EB_SUCCESS) {
      fprintf (stderr, "eb_set_appendix_subbook: %s\n",
	       eb_error_message(error_code));
      return 1;
    }
  }

  if (variables_get_bool(variables, EBU_USE_GAIJI_MAP, true)) {
    ebusub->gaiji = ebu_load_gaiji_map(ebu->book);
  }

  /* Use cp51932 to unicode map */
  EB_Character_Code charcode;
  eb_character_code(ebu->book, &charcode);
  if (charcode == EB_CHARCODE_JISX0208_GB2312 ||
      charcode == EB_CHARCODE_JISX0208) {
    ebusub->gaiji = ebu_gaiji_register_cp51932(ebusub->gaiji);
  }

  ebusub->uribase->used = 0;
  sbuffer_write_snprintf(ebusub->uribase, "eb:%s/%%d:%%d", ebusub->name);

  EB_Position pos;
  if (eb_menu(ebu->book, &pos) == EB_SUCCESS) {
    printf ("Registering menu\n");
    eb_seek_text (ebu->book, &pos);
    ebu_register_content0(ebu, ebusub);
  }

  if (eb_copyright(ebu->book, &pos) == EB_SUCCESS) {
    printf ("Registering copyright\n");
    eb_seek_text (ebu->book, &pos);
    ebu_register_content0(ebu, ebusub);
  }

  if (eb_text(ebu->book, &pos) == EB_SUCCESS) {
    printf ("Registering text\n");
    eb_seek_text (ebu->book, &pos);
    ebu_register_content0(ebu, ebusub);
  }

  if (ebusub->gaiji) {
    kvs_destroy(ebusub->gaiji);
    ebusub->gaiji = NULL;
  }

  return 0;
}

int ebu_make_index (EBU_Params *ebu) {
  int count;
  EB_Subbook_Code subbook_list[EB_MAX_SUBBOOKS];
  unsigned char names[EB_MAX_SUBBOOKS][EB_MAX_DIRECTORY_NAME_LENGTH + 3 + 1];
  EBUSUB *ebusub = ebusub_create();
  if (!ebusub) return 1;
  
  set_signal_handler();

  est_db_add_attr_index(ebu->estdb, "eblook_code", ESTIDXATTRNUM);
  EB_Error_Code error_code = eb_subbook_list(ebu->book, subbook_list, &count);
  if (error_code != EB_SUCCESS) {
    fprintf (stderr, "eb_subbook_list(): %s\n", eb_error_message(error_code));
    return 1;
  }
  printf ("Number of subbooks is %d.\n", count);

  for (int i = 0; i < count; i++) {
    eb_subbook_directory2(ebu->book, subbook_list[i], names[i]);
    if (error_code != EB_SUCCESS) {
      fprintf (stderr, "eb_subbook_directory2(): %s\n",
	       eb_error_message(error_code));
      return 1;
    }

    for (int j = 0; j < i; j++) {
      if (!strcmp(names[j], names[i])) {
	sprintf(names[i] + strlen(names[i]), ".%d", i + 1);
	break;
      }
    }

    fprintf(stderr, "name is %s\n", names[i]);
    ebusub->name = names[i];
    ebusub->code = subbook_list[i];

    if (ebu_make_index_subbook(ebu, ebusub)) {
      fprintf (stderr, "ebu_make_index_subbook() failed\n");
      return 1;
    }
  }

  ebusub->name = NULL;
  ebusub_destroy(ebusub);

  return 0;
}
