/*
  Copyright (c) 2022  Kazuhiro Ito (kzhr@d1.dion.ne.jp)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <strings.h>
#include <dirent.h>
#include <iconv.h>
#include <ebu/eb.h>
#include <ebu/text.h>
#include <ebu/font.h>
#include <ebu/appendix.h>
#include <ebu/error.h>
#include <ebu/binary.h>

#include "common.h"

static const unsigned char *cp51932_table[] =
  { "ada1", "\xe2\x91\xa0",
    "ada2", "\xe2\x91\xa1",
    "ada3", "\xe2\x91\xa2",
    "ada4", "\xe2\x91\xa3",
    "ada5", "\xe2\x91\xa4",
    "ada6", "\xe2\x91\xa5",
    "ada7", "\xe2\x91\xa6",
    "ada8", "\xe2\x91\xa7",
    "ada9", "\xe2\x91\xa8",
    "adaa", "\xe2\x91\xa9",
    "adab", "\xe2\x91\xaa",
    "adac", "\xe2\x91\xab",
    "adad", "\xe2\x91\xac",
    "adae", "\xe2\x91\xad",
    "adaf", "\xe2\x91\xae",
    "adb0", "\xe2\x91\xaf",
    "adb1", "\xe2\x91\xb0",
    "adb2", "\xe2\x91\xb1",
    "adb3", "\xe2\x91\xb2",
    "adb4", "\xe2\x91\xb3",
    "adb5", "\xe2\x85\xa0",
    "adb6", "\xe2\x85\xa1",
    "adb7", "\xe2\x85\xa2",
    "adb8", "\xe2\x85\xa3",
    "adb9", "\xe2\x85\xa4",
    "adba", "\xe2\x85\xa5",
    "adbb", "\xe2\x85\xa6",
    "adbc", "\xe2\x85\xa7",
    "adbd", "\xe2\x85\xa8",
    "adbe", "\xe2\x85\xa9",
    "adc0", "\xe3\x8d\x89",
    "adc1", "\xe3\x8c\x94",
    "adc2", "\xe3\x8c\xa2",
    "adc3", "\xe3\x8d\x8d",
    "adc4", "\xe3\x8c\x98",
    "adc5", "\xe3\x8c\xa7",
    "adc6", "\xe3\x8c\x83",
    "adc7", "\xe3\x8c\xb6",
    "adc8", "\xe3\x8d\x91",
    "adc9", "\xe3\x8d\x97",
    "adca", "\xe3\x8c\x8d",
    "adcb", "\xe3\x8c\xa6",
    "adcc", "\xe3\x8c\xa3",
    "adcd", "\xe3\x8c\xab",
    "adce", "\xe3\x8d\x8a",
    "adcf", "\xe3\x8c\xbb",
    "add0", "\xe3\x8e\x9c",
    "add1", "\xe3\x8e\x9d",
    "add2", "\xe3\x8e\x9e",
    "add3", "\xe3\x8e\x8e",
    "add4", "\xe3\x8e\x8f",
    "add5", "\xe3\x8f\x84",
    "add6", "\xe3\x8e\xa1",
    "addf", "\xe3\x8d\xbb",
    "ade0", "\xe3\x80\x9d",
    "ade1", "\xe3\x80\x9f",
    "ade2", "\xe2\x84\x96",
    "ade3", "\xe3\x8f\x8d",
    "ade4", "\xe2\x84\xa1",
    "ade5", "\xe3\x8a\xa4",
    "ade6", "\xe3\x8a\xa5",
    "ade7", "\xe3\x8a\xa6",
    "ade8", "\xe3\x8a\xa7",
    "ade9", "\xe3\x8a\xa8",
    "adea", "\xe3\x88\xb1",
    "adeb", "\xe3\x88\xb2",
    "adec", "\xe3\x88\xb9",
    "aded", "\xe3\x8d\xbe",
    "adee", "\xe3\x8d\xbd",
    "adef", "\xe3\x8d\xbc",
    "adf0", "\xe2\x89\x92",
    "adf1", "\xe2\x89\xa1",
    "adf2", "\xe2\x88\xab",
    "adf3", "\xe2\x88\xae",
    "adf4", "\xe2\x88\x91",
    "adf5", "\xe2\x88\x9a",
    "adf6", "\xe2\x8a\xa5",
    "adf7", "\xe2\x88\xa0",
    "adf8", "\xe2\x88\x9f",
    "adf9", "\xe2\x8a\xbf",
    "adfa", "\xe2\x88\xb5",
    "adfb", "\xe2\x88\xa9",
    "adfc", "\xe2\x88\xaa",
    "f9a1", "\xe7\xba\x8a",
    "f9a2", "\xe8\xa4\x9c",
    "f9a3", "\xe9\x8d\x88",
    "f9a4", "\xe9\x8a\x88",
    "f9a5", "\xe8\x93\x9c",
    "f9a6", "\xe4\xbf\x89",
    "f9a7", "\xe7\x82\xbb",
    "f9a8", "\xe6\x98\xb1",
    "f9a9", "\xe6\xa3\x88",
    "f9aa", "\xe9\x8b\xb9",
    "f9ab", "\xe6\x9b\xbb",
    "f9ac", "\xe5\xbd\x85",
    "f9ad", "\xe4\xb8\xa8",
    "f9ae", "\xe4\xbb\xa1",
    "f9af", "\xe4\xbb\xbc",
    "f9b0", "\xe4\xbc\x80",
    "f9b1", "\xe4\xbc\x83",
    "f9b2", "\xe4\xbc\xb9",
    "f9b3", "\xe4\xbd\x96",
    "f9b4", "\xe4\xbe\x92",
    "f9b5", "\xe4\xbe\x8a",
    "f9b6", "\xe4\xbe\x9a",
    "f9b7", "\xe4\xbe\x94",
    "f9b8", "\xe4\xbf\x8d",
    "f9b9", "\xe5\x81\x80",
    "f9ba", "\xe5\x80\xa2",
    "f9bb", "\xe4\xbf\xbf",
    "f9bc", "\xe5\x80\x9e",
    "f9bd", "\xe5\x81\x86",
    "f9be", "\xe5\x81\xb0",
    "f9bf", "\xe5\x81\x82",
    "f9c0", "\xe5\x82\x94",
    "f9c1", "\xe5\x83\xb4",
    "f9c2", "\xe5\x83\x98",
    "f9c3", "\xe5\x85\x8a",
    "f9c4", "\xe5\x85\xa4",
    "f9c5", "\xe5\x86\x9d",
    "f9c6", "\xe5\x86\xbe",
    "f9c7", "\xe5\x87\xac",
    "f9c8", "\xe5\x88\x95",
    "f9c9", "\xe5\x8a\x9c",
    "f9ca", "\xe5\x8a\xa6",
    "f9cb", "\xe5\x8b\x80",
    "f9cc", "\xe5\x8b\x9b",
    "f9cd", "\xe5\x8c\x80",
    "f9ce", "\xe5\x8c\x87",
    "f9cf", "\xe5\x8c\xa4",
    "f9d0", "\xe5\x8d\xb2",
    "f9d1", "\xe5\x8e\x93",
    "f9d2", "\xe5\x8e\xb2",
    "f9d3", "\xe5\x8f\x9d",
    "f9d4", "\xef\xa8\x8e",
    "f9d5", "\xe5\x92\x9c",
    "f9d6", "\xe5\x92\x8a",
    "f9d7", "\xe5\x92\xa9",
    "f9d8", "\xe5\x93\xbf",
    "f9d9", "\xe5\x96\x86",
    "f9da", "\xe5\x9d\x99",
    "f9db", "\xe5\x9d\xa5",
    "f9dc", "\xe5\x9e\xac",
    "f9dd", "\xe5\x9f\x88",
    "f9de", "\xe5\x9f\x87",
    "f9df", "\xef\xa8\x8f",
    "f9e0", "\xef\xa8\x90",
    "f9e1", "\xe5\xa2\x9e",
    "f9e2", "\xe5\xa2\xb2",
    "f9e3", "\xe5\xa4\x8b",
    "f9e4", "\xe5\xa5\x93",
    "f9e5", "\xe5\xa5\x9b",
    "f9e6", "\xe5\xa5\x9d",
    "f9e7", "\xe5\xa5\xa3",
    "f9e8", "\xe5\xa6\xa4",
    "f9e9", "\xe5\xa6\xba",
    "f9ea", "\xe5\xad\x96",
    "f9eb", "\xe5\xaf\x80",
    "f9ec", "\xe7\x94\xaf",
    "f9ed", "\xe5\xaf\x98",
    "f9ee", "\xe5\xaf\xac",
    "f9ef", "\xe5\xb0\x9e",
    "f9f0", "\xe5\xb2\xa6",
    "f9f1", "\xe5\xb2\xba",
    "f9f2", "\xe5\xb3\xb5",
    "f9f3", "\xe5\xb4\xa7",
    "f9f4", "\xe5\xb5\x93",
    "f9f5", "\xef\xa8\x91",
    "f9f6", "\xe5\xb5\x82",
    "f9f7", "\xe5\xb5\xad",
    "f9f8", "\xe5\xb6\xb8",
    "f9f9", "\xe5\xb6\xb9",
    "f9fa", "\xe5\xb7\x90",
    "f9fb", "\xe5\xbc\xa1",
    "f9fc", "\xe5\xbc\xb4",
    "f9fd", "\xe5\xbd\xa7",
    "f9fe", "\xe5\xbe\xb7",
    "faa1", "\xe5\xbf\x9e",
    "faa2", "\xe6\x81\x9d",
    "faa3", "\xe6\x82\x85",
    "faa4", "\xe6\x82\x8a",
    "faa5", "\xe6\x83\x9e",
    "faa6", "\xe6\x83\x95",
    "faa7", "\xe6\x84\xa0",
    "faa8", "\xe6\x83\xb2",
    "faa9", "\xe6\x84\x91",
    "faaa", "\xe6\x84\xb7",
    "faab", "\xe6\x84\xb0",
    "faac", "\xe6\x86\x98",
    "faad", "\xe6\x88\x93",
    "faae", "\xe6\x8a\xa6",
    "faaf", "\xe6\x8f\xb5",
    "fab0", "\xe6\x91\xa0",
    "fab1", "\xe6\x92\x9d",
    "fab2", "\xe6\x93\x8e",
    "fab3", "\xe6\x95\x8e",
    "fab4", "\xe6\x98\x80",
    "fab5", "\xe6\x98\x95",
    "fab6", "\xe6\x98\xbb",
    "fab7", "\xe6\x98\x89",
    "fab8", "\xe6\x98\xae",
    "fab9", "\xe6\x98\x9e",
    "faba", "\xe6\x98\xa4",
    "fabb", "\xe6\x99\xa5",
    "fabc", "\xe6\x99\x97",
    "fabd", "\xe6\x99\x99",
    "fabe", "\xef\xa8\x92",
    "fabf", "\xe6\x99\xb3",
    "fac0", "\xe6\x9a\x99",
    "fac1", "\xe6\x9a\xa0",
    "fac2", "\xe6\x9a\xb2",
    "fac3", "\xe6\x9a\xbf",
    "fac4", "\xe6\x9b\xba",
    "fac5", "\xe6\x9c\x8e",
    "fac6", "\xef\xa4\xa9",
    "fac7", "\xe6\x9d\xa6",
    "fac8", "\xe6\x9e\xbb",
    "fac9", "\xe6\xa1\x92",
    "faca", "\xe6\x9f\x80",
    "facb", "\xe6\xa0\x81",
    "facc", "\xe6\xa1\x84",
    "facd", "\xe6\xa3\x8f",
    "face", "\xef\xa8\x93",
    "facf", "\xe6\xa5\xa8",
    "fad0", "\xef\xa8\x94",
    "fad1", "\xe6\xa6\x98",
    "fad2", "\xe6\xa7\xa2",
    "fad3", "\xe6\xa8\xb0",
    "fad4", "\xe6\xa9\xab",
    "fad5", "\xe6\xa9\x86",
    "fad6", "\xe6\xa9\xb3",
    "fad7", "\xe6\xa9\xbe",
    "fad8", "\xe6\xab\xa2",
    "fad9", "\xe6\xab\xa4",
    "fada", "\xe6\xaf\x96",
    "fadb", "\xe6\xb0\xbf",
    "fadc", "\xe6\xb1\x9c",
    "fadd", "\xe6\xb2\x86",
    "fade", "\xe6\xb1\xaf",
    "fadf", "\xe6\xb3\x9a",
    "fae0", "\xe6\xb4\x84",
    "fae1", "\xe6\xb6\x87",
    "fae2", "\xe6\xb5\xaf",
    "fae3", "\xe6\xb6\x96",
    "fae4", "\xe6\xb6\xac",
    "fae5", "\xe6\xb7\x8f",
    "fae6", "\xe6\xb7\xb8",
    "fae7", "\xe6\xb7\xb2",
    "fae8", "\xe6\xb7\xbc",
    "fae9", "\xe6\xb8\xb9",
    "faea", "\xe6\xb9\x9c",
    "faeb", "\xe6\xb8\xa7",
    "faec", "\xe6\xb8\xbc",
    "faed", "\xe6\xba\xbf",
    "faee", "\xe6\xbe\x88",
    "faef", "\xe6\xbe\xb5",
    "faf0", "\xe6\xbf\xb5",
    "faf1", "\xe7\x80\x85",
    "faf2", "\xe7\x80\x87",
    "faf3", "\xe7\x80\xa8",
    "faf4", "\xe7\x82\x85",
    "faf5", "\xe7\x82\xab",
    "faf6", "\xe7\x84\x8f",
    "faf7", "\xe7\x84\x84",
    "faf8", "\xe7\x85\x9c",
    "faf9", "\xe7\x85\x86",
    "fafa", "\xe7\x85\x87",
    "fafb", "\xef\xa8\x95",
    "fafc", "\xe7\x87\x81",
    "fafd", "\xe7\x87\xbe",
    "fafe", "\xe7\x8a\xb1",
    "fba1", "\xe7\x8a\xbe",
    "fba2", "\xe7\x8c\xa4",
    "fba3", "\xef\xa8\x96",
    "fba4", "\xe7\x8d\xb7",
    "fba5", "\xe7\x8e\xbd",
    "fba6", "\xe7\x8f\x89",
    "fba7", "\xe7\x8f\x96",
    "fba8", "\xe7\x8f\xa3",
    "fba9", "\xe7\x8f\x92",
    "fbaa", "\xe7\x90\x87",
    "fbab", "\xe7\x8f\xb5",
    "fbac", "\xe7\x90\xa6",
    "fbad", "\xe7\x90\xaa",
    "fbae", "\xe7\x90\xa9",
    "fbaf", "\xe7\x90\xae",
    "fbb0", "\xe7\x91\xa2",
    "fbb1", "\xe7\x92\x89",
    "fbb2", "\xe7\x92\x9f",
    "fbb3", "\xe7\x94\x81",
    "fbb4", "\xe7\x95\xaf",
    "fbb5", "\xe7\x9a\x82",
    "fbb6", "\xe7\x9a\x9c",
    "fbb7", "\xe7\x9a\x9e",
    "fbb8", "\xe7\x9a\x9b",
    "fbb9", "\xe7\x9a\xa6",
    "fbba", "\xef\xa8\x97",
    "fbbb", "\xe7\x9d\x86",
    "fbbc", "\xe5\x8a\xaf",
    "fbbd", "\xe7\xa0\xa1",
    "fbbe", "\xe7\xa1\x8e",
    "fbbf", "\xe7\xa1\xa4",
    "fbc0", "\xe7\xa1\xba",
    "fbc1", "\xe7\xa4\xb0",
    "fbc2", "\xef\xa8\x98",
    "fbc3", "\xef\xa8\x99",
    "fbc4", "\xef\xa8\x9a",
    "fbc5", "\xe7\xa6\x94",
    "fbc6", "\xef\xa8\x9b",
    "fbc7", "\xe7\xa6\x9b",
    "fbc8", "\xe7\xab\x91",
    "fbc9", "\xe7\xab\xa7",
    "fbca", "\xef\xa8\x9c",
    "fbcb", "\xe7\xab\xab",
    "fbcc", "\xe7\xae\x9e",
    "fbcd", "\xef\xa8\x9d",
    "fbce", "\xe7\xb5\x88",
    "fbcf", "\xe7\xb5\x9c",
    "fbd0", "\xe7\xb6\xb7",
    "fbd1", "\xe7\xb6\xa0",
    "fbd2", "\xe7\xb7\x96",
    "fbd3", "\xe7\xb9\x92",
    "fbd4", "\xe7\xbd\x87",
    "fbd5", "\xe7\xbe\xa1",
    "fbd6", "\xef\xa8\x9e",
    "fbd7", "\xe8\x8c\x81",
    "fbd8", "\xe8\x8d\xa2",
    "fbd9", "\xe8\x8d\xbf",
    "fbda", "\xe8\x8f\x87",
    "fbdb", "\xe8\x8f\xb6",
    "fbdc", "\xe8\x91\x88",
    "fbdd", "\xe8\x92\xb4",
    "fbde", "\xe8\x95\x93",
    "fbdf", "\xe8\x95\x99",
    "fbe0", "\xe8\x95\xab",
    "fbe1", "\xef\xa8\x9f",
    "fbe2", "\xe8\x96\xb0",
    "fbe3", "\xef\xa8\xa0",
    "fbe4", "\xef\xa8\xa1",
    "fbe5", "\xe8\xa0\x87",
    "fbe6", "\xe8\xa3\xb5",
    "fbe7", "\xe8\xa8\x92",
    "fbe8", "\xe8\xa8\xb7",
    "fbe9", "\xe8\xa9\xb9",
    "fbea", "\xe8\xaa\xa7",
    "fbeb", "\xe8\xaa\xbe",
    "fbec", "\xe8\xab\x9f",
    "fbed", "\xef\xa8\xa2",
    "fbee", "\xe8\xab\xb6",
    "fbef", "\xe8\xad\x93",
    "fbf0", "\xe8\xad\xbf",
    "fbf1", "\xe8\xb3\xb0",
    "fbf2", "\xe8\xb3\xb4",
    "fbf3", "\xe8\xb4\x92",
    "fbf4", "\xe8\xb5\xb6",
    "fbf5", "\xef\xa8\xa3",
    "fbf6", "\xe8\xbb\x8f",
    "fbf7", "\xef\xa8\xa4",
    "fbf8", "\xef\xa8\xa5",
    "fbf9", "\xe9\x81\xa7",
    "fbfa", "\xe9\x83\x9e",
    "fbfb", "\xef\xa8\xa6",
    "fbfc", "\xe9\x84\x95",
    "fbfd", "\xe9\x84\xa7",
    "fbfe", "\xe9\x87\x9a",
    "fca1", "\xe9\x87\x97",
    "fca2", "\xe9\x87\x9e",
    "fca3", "\xe9\x87\xad",
    "fca4", "\xe9\x87\xae",
    "fca5", "\xe9\x87\xa4",
    "fca6", "\xe9\x87\xa5",
    "fca7", "\xe9\x88\x86",
    "fca8", "\xe9\x88\x90",
    "fca9", "\xe9\x88\x8a",
    "fcaa", "\xe9\x88\xba",
    "fcab", "\xe9\x89\x80",
    "fcac", "\xe9\x88\xbc",
    "fcad", "\xe9\x89\x8e",
    "fcae", "\xe9\x89\x99",
    "fcaf", "\xe9\x89\x91",
    "fcb0", "\xe9\x88\xb9",
    "fcb1", "\xe9\x89\xa7",
    "fcb2", "\xe9\x8a\xa7",
    "fcb3", "\xe9\x89\xb7",
    "fcb4", "\xe9\x89\xb8",
    "fcb5", "\xe9\x8b\xa7",
    "fcb6", "\xe9\x8b\x97",
    "fcb7", "\xe9\x8b\x99",
    "fcb8", "\xe9\x8b\x90",
    "fcb9", "\xef\xa8\xa7",
    "fcba", "\xe9\x8b\x95",
    "fcbb", "\xe9\x8b\xa0",
    "fcbc", "\xe9\x8b\x93",
    "fcbd", "\xe9\x8c\xa5",
    "fcbe", "\xe9\x8c\xa1",
    "fcbf", "\xe9\x8b\xbb",
    "fcc0", "\xef\xa8\xa8",
    "fcc1", "\xe9\x8c\x9e",
    "fcc2", "\xe9\x8b\xbf",
    "fcc3", "\xe9\x8c\x9d",
    "fcc4", "\xe9\x8c\x82",
    "fcc5", "\xe9\x8d\xb0",
    "fcc6", "\xe9\x8d\x97",
    "fcc7", "\xe9\x8e\xa4",
    "fcc8", "\xe9\x8f\x86",
    "fcc9", "\xe9\x8f\x9e",
    "fcca", "\xe9\x8f\xb8",
    "fccb", "\xe9\x90\xb1",
    "fccc", "\xe9\x91\x85",
    "fccd", "\xe9\x91\x88",
    "fcce", "\xe9\x96\x92",
    "fccf", "\xef\xa7\x9c",
    "fcd0", "\xef\xa8\xa9",
    "fcd1", "\xe9\x9a\x9d",
    "fcd2", "\xe9\x9a\xaf",
    "fcd3", "\xe9\x9c\xb3",
    "fcd4", "\xe9\x9c\xbb",
    "fcd5", "\xe9\x9d\x83",
    "fcd6", "\xe9\x9d\x8d",
    "fcd7", "\xe9\x9d\x8f",
    "fcd8", "\xe9\x9d\x91",
    "fcd9", "\xe9\x9d\x95",
    "fcda", "\xe9\xa1\x97",
    "fcdb", "\xe9\xa1\xa5",
    "fcdc", "\xef\xa8\xaa",
    "fcdd", "\xef\xa8\xab",
    "fcde", "\xe9\xa4\xa7",
    "fcdf", "\xef\xa8\xac",
    "fce0", "\xe9\xa6\x9e",
    "fce1", "\xe9\xa9\x8e",
    "fce2", "\xe9\xab\x99",
    "fce3", "\xe9\xab\x9c",
    "fce4", "\xe9\xad\xb5",
    "fce5", "\xe9\xad\xb2",
    "fce6", "\xe9\xae\x8f",
    "fce7", "\xe9\xae\xb1",
    "fce8", "\xe9\xae\xbb",
    "fce9", "\xe9\xb0\x80",
    "fcea", "\xe9\xb5\xb0",
    "fceb", "\xe9\xb5\xab",
    "fcec", "\xef\xa8\xad",
    "fced", "\xe9\xb8\x99",
    "fcee", "\xe9\xbb\x91",
    "fcf1", "\xe2\x85\xb0",
    "fcf2", "\xe2\x85\xb1",
    "fcf3", "\xe2\x85\xb2",
    "fcf4", "\xe2\x85\xb3",
    "fcf5", "\xe2\x85\xb4",
    "fcf6", "\xe2\x85\xb5",
    "fcf7", "\xe2\x85\xb6",
    "fcf8", "\xe2\x85\xb7",
    "fcf9", "\xe2\x85\xb8",
    "fcfa", "\xe2\x85\xb9",
    "fcfb", "\xef\xbf\xa2",
    "fcfc", "\xef\xbf\xa4",
    "fcfd", "\xef\xbc\x87",
    "fcfe", "\xef\xbc\x82"};

static void downcase (char * str) {
  for (int i = 0; str[i] != 0; i ++) {
    if (str[i] >= 'A' && str[i] <= 'Z') {
      str[i] += 'a' - 'A';
    }
  }
}

KVS *ebu_gaiji_register_cp51932(KVS *kvs) {
  DEBUG("Registering CP5132 translation table");
  if (!kvs) kvs = kvs_create();
  if (!kvs) goto fail;

  int count = sizeof(cp51932_table) / sizeof(cp51932_table[0]);
  DEBUG("Entry count is %d", count / 2);

  char buffer[1 + 4 + 1];
  for (int i = 0; i < count; i += 2) {
    sprintf(buffer, "e%s", cp51932_table[i]);
    if (kvs_set(kvs, buffer, cp51932_table[i + 1]))
      goto fail;
  }
  DEBUG("Registered");
  return kvs;

 fail:
  DEBUG("Failed to register");
  return kvs;
}

static SBuffer *ebu_gaiji_parse_replace(unsigned char *string) {
  DEBUG("ebu_gaiji_parse_replace(): Inupt is %s", string);
  SBuffer *buffer = sbuffer_create(4 * 3 + 1);
  int code;
  unsigned char *pos = string;
  char *p;

  while (*pos != 0) {
    if (*pos != 'u') goto error;

    code = strtol(pos + 1, &p, 16);
    pos = p;
    switch (*p) {
    case ',':
      pos++;
    case 0:
      break;
    default:
      fprintf(stderr, "not hex chars, %x\n", *p);
      goto error;
    }

    if (code  > 0x10ffff) {
      fprintf(stderr, "Illegal unicode characer, %s\n", string);
      goto error;
    }

    sbuffer_write_utf8(buffer, code);
  }
  sbuffer_request(buffer, 1);
  buffer->data[buffer->used] = 0;
  buffer->used++;

  /* fprintf(stderr, "replace is %s\n", buffer->data); */

  return buffer;

 error:
  return NULL;
}

KVS *ebu_gaiji_map_parse(SBuffer *indata) {
  size_t pos = 0;
  KVS *map = kvs_create();

  while (pos + 1 < indata->used) {
    if (indata->data[pos] == '#') {
      char *eol = strchr(indata->data + pos, '\n');
      if (eol) {
	pos = eol - (char *)indata->data + 1;
	continue;
      } else {
	/* EOF */
	break;
      }
    }

    char *eol = strchr(indata->data + pos, '\n');
    if (eol) *eol = 0;

    char *end_code = strchr(indata->data + pos, '\t');
    if (!end_code) {
      pos = eol - (char *)indata->data + 1;
      continue;
    }

    *end_code = 0;
    downcase(indata->data + pos);

    char *end_char = strchr(end_code + 1, '\t');
    if (end_char) *end_char = 0;
    /* for CRLF eol */
    end_char = strchr(end_code + 1, '\r');
    if (end_char) *end_char = 0;

    SBuffer * value = ebu_gaiji_parse_replace(end_code + 1);

    if (value) {
      VERBOSE("Gaiji %s will be replaces with %s",
	      indata->data + pos, value->data);
      kvs_set(map, indata->data + pos, value->data);
      sbuffer_destroy(value);
    }

    pos = eol - (char *)indata->data + 1;
  }

  fprintf(stdout, "gaiji map has %d entries\n", map->size);
  return map;
}

KVS *ebu_load_gaiji_map (EB_Book *book) {
  char path[EB_MAX_PATH_LENGTH + 1];
  char name[EB_MAX_DIRECTORY_NAME_LENGTH + 1];
  DIR *dir = NULL;

  eb_path(book, path);
  eb_subbook_directory(book, name);

  char *mapfilename = snprintf_buffer("%s.map", name);
  if (!mapfilename) goto fail;

  struct dirent *entry;

  dir = opendir(path);
  if (!dir) goto fail;

  do {
    entry = readdir(dir);
    if (!entry) goto fail;
  } while (strcasecmp(entry->d_name, mapfilename));

  free(mapfilename);
  mapfilename = snprintf_buffer("%s/%s", path, entry->d_name);
  if (!mapfilename) goto fail;

  VERBOSE("Loading gaiji map file: %s", mapfilename);
  SBuffer *mapdata = sbuffer_load_file(mapfilename);
  if (!mapdata) {
    fprintf(stderr, "Failed to load gaiji map file\n");
    goto fail;
  }
  KVS* result = ebu_gaiji_map_parse(mapdata);
  VERBOSE("Register %d gaiji replacements", result ? result->size : 0);

  sbuffer_destroy(mapdata);
  free(mapfilename);
  closedir (dir);
  return result;

 fail:
  if (mapfilename) free(mapfilename);
  if (dir) closedir (dir);
  return NULL;
}
