/*
  Copyright (c) 2022  Kazuhiro Ito (kzhr@d1.dion.ne.jp)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"

#define KVS_ENLARGE_STEP (16)

KVS* kvs_create() {
  KVS *result = malloc(sizeof(KVS));

  if (result) {
    result->size = 0;
    result->msize = 0;
    result->list = NULL;
  }

  return result;
}

void kvs_destroy(KVS *kvs) {
  if (kvs->list) {
    for (int i = 0; i < kvs->size; i++) {
      free(kvs->list[i].key);
      if (kvs->list[i].value) free(kvs->list[i].value);
    }
    free(kvs->list);
  }
  free(kvs);
}

static int kvs_expand(KVS* kvs, size_t step) {
  KVS_Element *ptr;
  size_t new_size = kvs->msize + step;

  if (kvs->list) {
    ptr = realloc(kvs->list, sizeof(KVS_Element) * new_size);
  } else {
    ptr = malloc(sizeof(KVS_Element) * new_size);
  }

  if (!ptr) return -1;

  kvs->list = ptr;
  kvs->msize = new_size;

  return 0;
}

/*
  Negative return value indicates insertion position.
  Insertion position is (-1 - return_value).
*/
static int kvs_search_pos(KVS *kvs, const unsigned char *key) {
  if (kvs->size == 0) return -1;

  int cur = kvs->size / 2;
  int min = -1, max = kvs->size;
  int r;

  while ((r = strcmp(key, kvs->list[cur].key)) != 0) {
    if (r > 0) {
      min = cur;
    } else {
      max = cur;
      if (max == 0) return -1;
    }

    cur = (max + min) / 2;
    if (cur == min) return (-2 - cur);
  }
  return cur;
}

unsigned char* kvs_get(KVS *kvs, const unsigned char *key) {
  int pos = kvs_search_pos(kvs, key);

  if (pos < 0) return NULL;
  return kvs->list[pos].value;
}

int kvs_unset(KVS *kvs, const unsigned char *key) {
  int pos = kvs_search_pos(kvs, key);

  if (pos >= 0) {
    if (kvs->list[pos].value) {
      free(kvs->list[pos].value);
      kvs->list[pos].value = NULL;
    }
  }
  return 0;
}

static int kvs_insert
(KVS *kvs, const unsigned char *key, const unsigned char *value, int pos) {
  int result;

  if (kvs->size + 1 > kvs->msize &&
      kvs_expand(kvs, KVS_ENLARGE_STEP)) {
    fprintf(stderr, "kvs_inset(): expansion failed\n");
    return -1;
  }

  unsigned char *new_key = strdup(key);
  unsigned char *new_value = strdup(value);
  if (!new_key || !new_value) {
    free(new_key);
    free(new_value);;
    return -1;
  }

  if (kvs->size && pos < kvs->size) {
    memmove (kvs->list + pos + 1, kvs->list + pos,
	     (kvs->size - pos) * sizeof (KVS_Element));
  }

  kvs->list[pos].key = new_key;
  kvs->list[pos].value = new_value;
  kvs->size ++;

  return 0;
}

int kvs_set(KVS *kvs, const unsigned char *key, const unsigned char *value) {
  if (!value || value[0] == 0) return kvs_unset(kvs, key);

  int pos = kvs_search_pos(kvs, key);
  if (pos < 0) {
    return kvs_insert(kvs, key, value, -1 - pos);
  } else {
    unsigned char *str;
    str = strdup(value);
    if (str) {
      if (kvs->list[pos].value) {
	free(kvs->list[pos].value);
      }
      kvs->list[pos].value = str;
      return 0;
    } else {
      return -1;
    }
  }
}

